-- Puffin: A Tojaso FowlPlay Project
-- Adapter component for creating and updating lists of spells tracking active auras or cooldowns.
-- Each list has an associated icon or bar group view which gets configured during initialization.
-- Lists are filtered, sorted and annotated in response to changes in models for unit auras and cooldowns.
-- Icon and bar group views are updated when changed are tracked in associated spell lists.

-- Spell attributes in a spell list:
-- name - spell name (text)
-- label - label to use for this spell (default: nil, just use the spell name)
-- id - spell id (number)
-- count - spell count (text)
-- expire - time at which the spell will expire (number)
-- duration - overall time the spell will last (number)
-- maxTime - duration to display (number, default: use spell duration)
-- icon - icon associated with the spell (texture)
-- r, g, b - color associated with the spell
-- source - source of the spell (e.g., "buffs", "cooldowns")
-- tooltip - source-dependent identifier used to display a tooltip
-- unit - unit that the spell is on
-- caster - name of unit that cast the spell

local Adapter_SpellTrackers = {}
local EF = Puffin.ErrorFormat
local DB, UN, IG, DP, AM, LSM, _
local active_groups = {}
local configuration_changed = true
local units_aura_tracked = {}
local units_aura_changed = {}
local cooldowns_tracked = false
local cooldowns_changed = false

-- Track aura changes to units associated with icon groups
local function UnitChanged(e, unit, why) if why == "auras" or why == "new" or why == "released" or why == "possess" then units_aura_changed[unit] = true end end

-- Track cooldown changes to support icon groups that display spells on cooldown
local function CooldownsChanged(e) cooldowns_changed = true end

-- Check an attribute to see if it refers to special media
local function CheckMedia(v)
	if type(v) == "table" and v[1] == "#LSM" then v = LSM:Fetch(v[2], v[3]) end -- check for LibSharedMedia references
	return v
end

-- Handle mouse interactions with views representing spells in a spell list
local function Spell_OnEnter(frame, eventHandler)
	local e = eventHandler; if not e then return end
	local tt, id, unit, caster, ttanchor = e.source, e.tooltip, e.unit, e.caster, e.ttanchor
	if tt and id and unit then
		if (not ttanchor or ttanchor == "DEFAULT") and (GetCVar("UberTooltips") == "1") then
			GameTooltip_SetDefaultAnchor(GameTooltip, frame)
		else
			if not ttanchor or (ttanchor == "DEFAULT") then ttanchor = "ANCHOR_BOTTOMLEFT" else ttanchor = "ANCHOR_" .. ttanchor end
			GameTooltip:SetOwner(frame, ttanchor)
		end
		if tt == "text" then
			GameTooltip:SetText(id)
		elseif (tt == "inventory") then
			local slot = GetInventorySlotInfo(id); if slot then GameTooltip:SetInventoryItem("player", slot) end
		elseif (tt == "spell link") or (tt == "item link") then
			GameTooltip:SetHyperlink(id)
		elseif (tt == "spell id") or (tt == "internal") then
			GameTooltip:SetSpellByID(id)
		elseif tt == "buffs" then
			GameTooltip:SetUnitBuff(unit, id)
		elseif tt == "debuffs" then
			GameTooltip:SetUnitDebuff(unit, id)
		elseif tt == "spell" then
			GameTooltip:SetText(id)
		elseif tt == "totem" then
			GameTooltip:SetTotem(id)
		end
		if caster and (caster ~= "") then GameTooltip:AddLine("<Applied by " .. caster .. ">", 0, 0.8, 1, false) end
		GameTooltip:Show()
	end
end

-- Handle mouse interactions with an icon within the view
local function Spell_OnLeave() GameTooltip:Hide() end

-- Handle mouse interactions with an icon within the view
local function Spell_OnClick(frame, eventHandler, button)
	local e = eventHandler; if not e then return end
	if not InCombatLockdown() and button == "RightButton" then
		local tt, id, unit = e.source, e.tooltip, e.unit
		if tt == "buffs" and id and unit == "player" then CancelUnitBuff(unit, id) end
	end
end

-- Check status for a group to see if it should be hidden
local function GroupStatus(gp)
	if InCinematic() or C_PetBattles.IsInBattle() then return true end
	return false
end

-- Apply filters to a spell from a given source to see if it should be displayed in the group
local function SpellFilter(gp, src, s)
	if src.whiteList and (not s.name or not src.whiteList[s.name]) then return end
	if src.blackList and s.name and src.blackList[s.name] then return false end
	if src.typeList then
	end
	if type(src.timerOptions) == "table" then
		for opt, value in pairs(src.timerOptions) do
			if opt == "maxDuration" and (s.duration == 0 or s.duration > value) then return false end
			if opt == "minDuration" and (s.duration > 0 and s.duration < value) then return false end
			if opt == "unlimitedDuration" and (value == true) ~= (s.duration == 0) then return false end
			if opt == "uniformDuration" then s.maxTime = value end
		end
	end
	return true
end

-- Apply sorters to the spells in a group, use simple bubble sort since spell lists are never long
-- Available sorters include: AtoZ, ZtoA, TimeUp, TimeDown, DurationUp, DurationDown, StartUp, StartDown, PlayerFirst, PlayerLast
local function GroupSorter(gp, spells)
	local sorters = gp.sorters
	local count = #spells
	local infinity = GetTime() + 144000
	if sorters and count > 1 then
		for i = 1, count - 1 do
			for j = i + 1, count do
				local k, kmax, flip, done, s1, s2 = 0, #sorters, false, false, spells[i], spells[j]
				while k < kmax and not done do
					k = k + 1
					local s = sorters[k]
					local v1, v2
					if s == "AtoZ" then v1 = s1.name; v2 = s2.name
					elseif s == "ZtoA" then v2 = s1.name; v1 = s2.name
					elseif s == "TimeUp" then v1 = s1.expire; v2 = s2.expire if v1 == 0 then v1 = infinity end if v2 == 0 then v2 = infinity end
					elseif s == "TimeDown" then v2 = s1.expire; v1 = s2.expire if v1 == 0 then v1 = infinity end if v2 == 0 then v2 = infinity end
					elseif s == "DurationUp" then v1 = s1.duration; v2 = s2.duration if v1 == 0 then v1 = 144000 end if v2 == 0 then v2 = 144000 end
					elseif s == "DurationDown" then v2 = s1.duration; v1 = s2.duration if v1 == 0 then v1 = 144000 end if v2 == 0 then v2 = 144000 end
					elseif s == "StartUp" then v1 = s1.expire - s1.duration; v2 = s2.expire - s2.duration
					elseif s == "StartDown" then v2 = s1.expire - s1.duration; v1 = s2.expire - s2.duration
					elseif s == "PlayerFirst" then v1 = s1.castByPlayer and 1 or 2; v2 = s2.castByPlayer and 1 or 2
					elseif s == "PlayerLast" then v1 = s1.castByPlayer and 2 or 1; v2 = s2.castByPlayer and 2 or 1
					end
					if v1 ~= v2 then flip = v1 > v2; done = true end
				end
				if flip then spells[i] = s2; spells[j] = s1 end
			end
		end
	end
end

-- Return RGB values of a color for a spell based on the current color scheme
local function GetColor(source, spell, atype, colorScheme, standardColors, spellColors)
	local c = nil
	if colorScheme == "player" and spell.castByPlayer then colorScheme = nil else c = standardColors.undefined end
	if not colorScheme or (colorScheme == "standard") then
		if atype and (spell.source == "debuffs") then c = standardColors[atype] else c = standardColors[source] end
	elseif colorScheme == "spell" then
		c = spellColors[spell.name or "none"]
	elseif colorScheme == "class" then
		if spell.class then c = RAID_CLASS_COLORS[spell.class]; if c then return c.r, c.g, c.b end end
	elseif not c and type(colorScheme) == "string" then
		c = colorScheme -- either unknown color scheme or color encoded as hex string
	end
	if not c then c = standardColors.default end
	local n = tonumber(c, 16) or 0
	return math.floor(n / (256 * 256)) / 255, (math.floor(n / 256) % 256) / 255, (n % 256) / 255
end

function Adapter_SpellTrackers.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("Adapter_SpellTrackers", "Data component not found"))
	UN = Puffin.Component("Model_Units")
	assert(UN, EF("Adapter_SpellTrackers", "Model_Units component not found"))
	IG = Puffin.Component("View_IconGroup")
	assert(IG, EF("Adapter_SpellTrackers", "View_IconGroup component not found"))
	BG = Puffin.Component("View_BarGroup")
	assert(BG, EF("Adapter_SpellTrackers", "View_BarGroup component not found"))
	AM = Puffin.Component("View_AnchorManagement")
	assert(AM, EF("Adapter_SpellTrackers", "View_AnchorManagement component not found"))
	LSM = LibStub("LibSharedMedia-3.0")
	assert(LSM, EF("Adapter_Configuration", "LibSharedMedia not found"))
end

-- On first update, configure the icon groups, afterwards just update each as needed
function Adapter_SpellTrackers.OnUpdate()
	local groups = DB.GetRecord("spell_trackers", "groups")
	if not groups then return end
	local spellColors = DB.GetRecord("defaults", "spell_colors")
	local standardColors = DB.GetRecord("defaults", "standard_colors")
	assert(spellColors and standardColors, EF("Adapter_SpellTrackers", "color tables not found"))

	if configuration_changed then -- can not configure during combat lockdown
		for name, gp in pairs(groups) do -- allocate and configure each group and associated view
			local layout = DB.GetRecord("layouts", gp.layout)
			if layout then
				local v = layout.view == "bars" and BG or IG -- determine which view component to use
				local u = v.Create() -- create the icon or bar group view
				local a = gp.anchor
				if a then
					u.offsetX = a.offsetX; u.offsetY = a.offsetY; u.anchorPoint = a.anchorPoint; u.relativePoint = a.relativePoint
					u.relativeTo = a.relativeTo; u.relativeEmpty = a.relativeEmpty; u.relativeX = a.relativeX; u.relativeY = a.relativeY
					u.relativeLastBar = a.relativeLastBar; u.relativeRow = a.relativeRow; u.relativeColumn = a.relativeColumn
				else u.offsetX = 0; u.offsetY = 0; u.anchorPoint = "CENTER"; u.relativePoint = "CENTER" end
				u.name = name; u.defaults = gp.defaults; u.layout = layout; u.sfx = gp.sfx
				u.depth = gp.depth; u.scale = gp.scale; u.alpha = gp.alpha; u.hidden = false
				u.onEnter = Spell_OnEnter; u.onLeave = Spell_OnLeave; u.onClick = Spell_OnClick
				v.Configure(u); AM.RegisterView(name, u)
				active_groups[name] = { component = v, view = u, group = gp, hidden = false, spells = {} }				
				if gp.sources then
					for _, src in ipairs(gp.sources) do
						local unit, data = src.unit, src.data
						if unit and (data == "buffs" or data == "debuffs" or data == "stance" or data == "power" or data == "weapon" or data == "missing") then
							if not units_aura_tracked[unit] then
								units_aura_tracked[unit] = true; units_aura_changed[unit] = true
								Puffin.RegisterCallback(UN.callbacks[unit], UnitChanged)
							end
						elseif data == "cooldowns" then
							if not cooldowns_tracked then
								cooldowns_tracked = true; cooldowns_changed = true
								Puffin.RegisterCallback("Cooldowns", CooldownsChanged)
							end
						end
					end
				end
			end
		end
		configuration_changed = false -- don't process icon groups first time through
		return
	end
	
	for name, group in pairs(active_groups) do -- update all icon and bar group views
		local gp = group.group
		if gp and gp.sources then
			local update_group, hide_group = false, GroupStatus(gp) -- check status to see if group is hidden
			if not hide_group or gp.hidden then
				for _, src in ipairs(gp.sources) do -- iterate through sources and see if need to update due to changes in model
					local unit, data = src.unit, src.data
					if unit and (data == "buffs" or data == "debuffs" or data == "stance" or data == "power" or data == "weapon" or data == "missing") then
						if units_aura_changed[unit] then update_group = true end
					elseif data == "cooldowns" then
						if cooldowns_changed then update_group = true end
					end
				end
				if hide_group or update_group then
					for _, s in ipairs(group.spells) do Puffin.ReleaseTable(s) end -- release previous spell list contents
					table.wipe(group.spells) -- reset spells to be displayed in the icon or bar group
					local count = 0
					local cs = gp.sfx and gp.sfx.color or nil -- color scheme for the group, if any
					if not hide_group then
						for _, src in ipairs(gp.sources) do -- iterate through sources, filter each spell before adding to the list
							if src.unit and (src.data == "buffs" or src.data == "debuffs") then -- process aura-related sources
								local t = DB.GetRecord("units", src.unit)
								if t and t.inVehicle and t.vehicle then t = DB.GetRecord("units", t.vehicle) end
								if t and t.auras then
									local ap = t.auras[src.data]
									for i = 1, ap.count do
										local a = ap[i]
										local castCheck = not src.castBy or (a.caster and UnitIsUnit(src.castBy, a.caster))
										if castCheck and SpellFilter(gp, src, a) then -- apply filters to make sure spell should be included in the list
											local s = Puffin.AllocateTable()
											s.unit = src.unit; s.name = a.name; s.duration = a.duration; s.expire = a.expire; s.maxTime = a.maxTime or a.duration
											s.count = a.count; s.icon = a.icon; s.id = a.id; s.source = src.data; s.tooltip = a.index; s.caster = a.casterName
											if a.caster then _, s.class = UnitClass(a.caster); s.castByPlayer = (a.caster == "player") end
											s.r, s.g, s.b = GetColor(src.data, s, a.atype, cs, standardColors, spellColors)
											count = count + 1; group.spells[count] = s
										end
									end
								end
							elseif src.data == "cooldowns" then
								for cname, cd in DB.GetRecords("cooldowns") do
									local unitCheck = not src.unit or (src.unit == cd.unit); cd.expire = cd.start + cd.duration
									if unitCheck and SpellFilter(gp, src, cd) then -- apply filters to make sure spell should be included in the list
										local s = Puffin.AllocateTable()
										s.unit = cd.unit; if cd.ttType == "spell link" then _, s.class = UnitClass(s.unit) end
										s.name = cname; s.duration = cd.duration; s.expire = cd.start + cd.duration; s.maxTime = cd.maxTime or cd.duration
										s.count = cd.count; s.icon = cd.icon; s.id = cd.id; s.source = cd.ttType; s.tooltip = cd.ttArg; s.castByPlayer = (src.unit == "player")
										if cd.count ~= nil then s.forceCount = true end
										s.r, s.g, s.b = GetColor(src.data, s, nil, cs, standardColors, spellColors)
										count = count + 1; group.spells[count] = s
									end
								end
							elseif src.data == "stance" then
								local a = DB.GetRecord("player", "stance")
								if a and a.name then
									local s = Puffin.AllocateTable()
									s.unit = "player"; _, s.class = UnitClass("player"); s.name = a.name; s.duration = 0; s.expire = 0
									s.count = 1; s.icon = a.icon; s.source = "spell link"; s.tooltip = a.link; s.castByPlayer = true
									s.r, s.g, s.b = GetColor(src.data, s, nil, cs, standardColors, spellColors)
									count = count + 1; group.spells[count] = s
								end
							elseif src.data == "power" then
								local a = DB.GetRecord("player", "power")
								if a and a.name then
									local s = Puffin.AllocateTable()
									s.unit = "player"; _, s.class = UnitClass("player"); s.name = a.name; s.duration = 0; s.expire = 0; s.id = a.id
									s.count = a.count; s.icon = a.icon; s.source = "spell link"; s.tooltip = a.link; s.castByPlayer = true
									s.r, s.g, s.b = GetColor(src.data, s, nil, cs, standardColors, spellColors)
									count = count + 1; group.spells[count] = s
								end
							elseif src.data == "weapon" then
								local a = DB.GetRecord("player", "mainhand")
								if a and a.name then
									local s = Puffin.AllocateTable()
									s.unit = "player"; _, s.class = UnitClass("player"); s.name = a.name; s.duration = 3600; s.expire = a.expire
									s.count = a.count; s.icon = a.icon; s.source = "inventory"; s.tooltip = "MainHandSlot"; s.castByPlayer = true
									s.r, s.g, s.b = GetColor(src.data, s, nil, cs, standardColors, spellColors)
									count = count + 1; group.spells[count] = s
								end
								local a = DB.GetRecord("player", "offhand")
								if a and a.name then
									local s = Puffin.AllocateTable()
									s.unit = "player"; _, s.class = UnitClass("player"); s.name = a.name; s.duration = 3600; s.expire = a.expire
									s.count = a.count; s.icon = a.icon; s.source = "inventory"; s.tooltip = "SecondaryHandSlot"; s.castByPlayer = true
									s.r, s.g, s.b = GetColor(src.data, s, nil, cs, standardColors, spellColors)
									count = count + 1; group.spells[count] = s
								end
							elseif src.data == "missing" then
								local t = DB.GetRecord("player", "missing")
								if t then
									for spell, icon in pairs(t) do
										local s = Puffin.AllocateTable()
										s.unit = "player"; _, s.class = UnitClass("player"); s.name = spell; s.duration = 0; s.expire = 0
										s.count = 1; s.icon = icon; s.source = "text"; s.tooltip = "Missing " .. spell
										s.r, s.g, s.b = 255, 0, 0
										count = count + 1; group.spells[count] = s
									end
								end
							end
						end
						GroupSorter(gp, group.spells) -- sort the spell list
					end
					-- print("spells", name, count, GetTime())
					-- for i, s in ipairs(group.spells) do print(i, s.name, s.duration, s.expire, s.count) end
					local v = group.component
					v.Update(group.view, count, group.spells)
				end
			end
		end
	end

	AM.Update() -- anchor management
	table.wipe(units_aura_changed); cooldowns_changed = false
end

Puffin.RegisterComponent("Adapter_SpellTrackers", Adapter_SpellTrackers, "ADAPTER", 3)
