-- Puffin: A Tojaso FowlPlay Project
-- Adapter component for assembling unit frames and updating them as needed.
-- Unit frames are assembled from configuration layout records. Different configurations
-- can be enabled for based on current player status (e.g., in solo, party, or raid).
-- Placement is based on named anchor points (eventually these will be moveable).
-- This component must avoid configuring while in global lockdown. Also is responsible for
-- triggering showing and hiding of the default Blizzard unit frames.

local Adapter_UnitFrames = {}
local EF = Puffin.ErrorFormat
local DB, UN, UF, DP, LSM
local active_panels = {}
local active_units = {}
local unit_changed = {}
local configuration_mode = false
local configuration_changed = true
local configuration_updated = false
local units_configured = false
local party_configured = false
local raid_configured = false
local raid_groups = nil
local raid_groups_max = 8 -- optionally set in configuration defaults database to maximum number of raid groups to display
local raid_members_configured = 0
local raid_members_group = {}
local raid_members_index = {}
local clicks_record = nil -- cache of click-casting bindings

local function UnitChanged(e, unit, why) unit_changed[unit] = true end
local function ConfigurationChanged() configuration_changed = true; configuration_updated = false end

-- Check an attribute to see if it refers to special media
local function CheckMedia(v)
	if type(v) == "table" and v[1] == "#LSM" then v = LSM:Fetch(v[2], v[3]) end -- check for LibSharedMedia references
	return v
end

function Adapter_UnitFrames.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("Adapter_UnitFrames", "Data component not found"))
	UN = Puffin.Component("Model_Units")
	assert(UN, EF("Adapter_UnitFrames", "Model_Units component not found"))
	UF = Puffin.Component("View_UnitFrame")
	assert(UF, EF("Adapter_UnitFrames", "View_UnitFrame component not found"))
	DP = Puffin.Component("View_DisplayPanel")
	assert(DP, EF("Adapter_UnitFrames", "View_DisplayPanel component not found"))
	LSM = LibStub("LibSharedMedia-3.0")
	assert(LSM, EF("Adapter_Configuration", "LibSharedMedia not found"))
	
	Puffin.RegisterCallback("GROUP_ROSTER_UPDATE", ConfigurationChanged)
	Puffin.RegisterCallback("PARTY_CONVERTED_TO_RAID", ConfigurationChanged)
	Puffin.RegisterCallback("ACTIVE_TALENT_GROUP_CHANGED", ConfigurationChanged)
	
	local n = DB.GetAttribute("defaults", "number", "raid_groups_max")
	if n then raid_groups_max = n end
end

-- Initialize a new unit
local function CreateUnit(unit, uid, v, pt, layout, isRaid)
	local u = UF.Create(isRaid) -- create the unit frame view as either "normal" or "raid"
	u.unit = uid; u.defaults = v.defaults or pt.defaults or layout.defaults; u.templates = v.templates or pt.templates or layout.templates; u.widgets = v.widgets
	u.height = v.height; u.width = v.width; u.depth = pt.depth; u.clique = v.clique; u.raidGroup = nil
	active_units[unit] = u
	unit_changed[uid] = true; Puffin.RegisterCallback(UN.callbacks[uid], UnitChanged)
	return u
end

-- Configure a party unit
local function ConfigurePartyUnit(pt, layout, i, isPlayer)
	local pname, tname = "player", "target"
	if not isPlayer then pname = "party" .. i; tname = "party" .. i .. "target" if pt.player then i = i + 1 end end
	local anchor, anchorX, anchorY, anchorPoint, relativePoint = pt.anchor and layout.anchors[pt.anchor] or nil, 0, 0, nil, nil
	if anchor then
		anchorX = (i - 1) * (pt.offsetX or 0) + (anchor.x or 0); anchorY = (i - 1) * (pt.offsetY or 0) + (anchor.y or 0)
		anchorPoint = anchor.anchorPoint; relativePoint = anchor.relativePoint
	end
	if pt.party then
		local unit = isPlayer and "partyplayer" or pname
		local u = CreateUnit(unit, pname, pt.party, pt, layout, false)
		u.offsetX = pt.party.offsetX + anchorX; u.offsetY = pt.party.offsetY + anchorY
		u.anchorPoint = anchorPoint; u.relativePoint = relativePoint
		UF.Configure(u)
	end
	if pt.target then
		local unit = isPlayer and "partytarget" or tname
		local u = CreateUnit(unit, tname, pt.target, pt, layout, false)
		u.offsetX = pt.target.offsetX + anchorX; u.offsetY = pt.target.offsetY + anchorY
		u.anchorPoint = anchorPoint; u.relativePoint = relativePoint
		UF.Configure(u)
	end
end

-- Release special party player unit
local function ReleasePartyPlayer()
	local pname, tname = "partyplayer", "partytarget"
	local u = active_units[pname]; if u then Puffin.UnregisterCallback(UN.callbacks[u.unit], UnitChanged); UF.Release(u); active_units[pname] = nil end
	u = active_units[tname]; if u then Puffin.UnregisterCallback(UN.callbacks[u.unit], UnitChanged); UF.Release(u); active_units[tname] = nil end
end

-- Release a party unit
local function ReleasePartyUnit(i)
	local pname, tname = "party" .. i, "party" .. i .. "target"
	local u = active_units[pname]; if u then Puffin.UnregisterCallback(UN.callbacks[u.unit], UnitChanged); UF.Release(u); active_units[pname] = nil end
	u = active_units[tname]; if u then Puffin.UnregisterCallback(UN.callbacks[u.unit], UnitChanged); UF.Release(u); active_units[tname] = nil end
end

-- Configure a raid unit except for position
local function ConfigureRaidUnit(pt, layout, i)
	local pname, tname = "raid" .. i, "raid" .. i .. "target"
	if pt.player then CreateUnit(pname, pname, pt.player, pt, layout, true) end
	if pt.target then CreateUnit(tname, tname, pt.target, pt, layout, true) end
end

-- Position a raid unit and complete configuration
local function PositionRaidUnit(pt, layout, i, group, index)
	local pname, tname = "raid" .. i, "raid" .. i .. "target"
	local anchor, anchorX, anchorY, anchorPoint, relativePoint = pt.anchor and layout.anchors[pt.anchor] or nil, 0, 0, nil, nil
	if anchor then
		anchorX = (index * (pt.offsetX or 0)) + (anchor.x or 0) + ((group - 1) * (pt.groupOffsetX or 0))
		anchorY = (index * (pt.offsetY or 0)) + (anchor.y or 0) + ((group - 1) * (pt.groupOffsetY or 0))
		anchorPoint = anchor.anchorPoint; relativePoint = anchor.relativePoint
	end
	if pt.player then
		local u, v = active_units[pname], pt.player
		if u and v then
			u.offsetX = v.offsetX + anchorX; u.offsetY = v.offsetY + anchorY; u.raidGroup = group
			u.anchorPoint = anchorPoint; u.relativePoint = relativePoint
			UF.Configure(u)
		end
	end
	if pt.target then
		local u, v = active_units[tname], pt.target
		if u and v then
			u.offsetX = v.offsetX + anchorX; u.offsetY = v.offsetY + anchorY; u.raidGroup = group
			u.anchorPoint = anchorPoint; u.relativePoint = relativePoint
			UF.Configure(u)
		end
	end
end

-- Release a raid unit
local function ReleaseRaidUnit(i)
	local pname, tname = "raid" .. i, "raid" .. i .. "target"
	local u = active_units[pname]
	if u then UF.Release(u); active_units[pname] = nil; Puffin.UnregisterCallback(UN.callbacks[u.unit], UnitChanged) end
	u = active_units[tname]
	if u then UF.Release(u); active_units[tname] = nil; Puffin.UnregisterCallback(UN.callbacks[u.unit], UnitChanged) end
end

-- Configure a panel
local function ConfigurePanel(pt, layout, backdrop, units)
	local anchor, x, y, anchorPoint, relativePoint = pt.anchor and layout.anchors[pt.anchor] or nil, 0, 0, nil, nil
	if anchor then
		x = (backdrop.offsetX or 0) + (anchor.x or 0)
		y = (backdrop.offsetY or 0) + (anchor.y or 0)
		anchorPoint = anchor.anchorPoint; relativePoint = anchor.relativePoint
	end
	dp = DP.Create() -- create the backdrop panel
	dp.defaults = layout.defaults; dp.templates = backdrop.templates or pt.templates or layout.templates; dp.widgets = backdrop.widgets
	dp.height = backdrop.height; dp.width = backdrop.width; dp.depth = backdrop.depth or pt.depth or 0
	dp.offsetX = backdrop.offsetX + x; dp.offsetY = backdrop.offsetY + y
	dp.anchorPoint = anchorPoint; dp.relativePoint = relativePoint
	dp.unit = backdrop.unit
	if dp.unit then Puffin.RegisterCallback(UN.callbacks[dp.unit], UnitChanged) end
	DP.Configure(dp)
	active_panels[units] = dp
end

-- Release a panel
local function ReleasePanel(units)
	dp = active_panels[units]
	if dp then
		if dp.unit then Puffin.UnregisterCallback(UN.callbacks[dp.unit], UnitChanged) end
		DP.Release(dp); active_panels[units] = nil
	end
end

local function UF_OnChangeAttribute(self, attribute, value)
	if (attribute == "unit") and value then
		-- nothing to do here? yet clique loses track sometimes of frame changes... bug in clique? use method from Shadowed Unit Frames instead?
	end
end

local function UF_OnEnter(self)
	if GetCVar("UberTooltips") == "1" then GameTooltip_SetDefaultAnchor(GameTooltip, self) else GameTooltip:SetOwner(self, "BOTTOMRIGHT") end
	local unit = self:GetAttribute("unit")
	if unit then GameTooltip:SetUnit(unit) end
end

local function UF_OnLeave(self) GameTooltip:Hide() end

-- Initialize a newly created unit
local function InitializeUnitConfiguration(self, name)
	local frame = _G[name]
	if frame then
		frame:SetScript("OnAttributeChanged", UF_OnChangeAttribute)
		frame:SetScript("OnEnter", UF_OnEnter)
		frame:SetScript("OnLeave", UF_OnLeave)
		frame:RegisterForClicks("AnyUp")

		local header = frame:GetParent()
		if Clique and header then
			if header.UseClique and not frame._clique then
				Clique:RegisterFrame(frame); frame._clique = true
			elseif not header.UseClique and frame._clique then
				Clique:UnregisterFrame(frame); frame._clique = nil
			end
		elseif clicks_record then
			for _, t in pairs(clicks_record) do frame:SetAttribute(t.a, t.v) end
		end
	end
end

-- Securely create and initialize a raid unit
local CreateSecureUnit = [[
	local header = self:GetParent()
	local suffix = header:GetAttribute("SecureUnitSuffix")
	if suffix then self:SetAttribute("unitsuffix", suffix) end
	local w, h = header:GetAttribute("initial-width"), header:GetAttribute("initial-height")
	self:SetWidth(w); self:SetHeight(h)
	self:SetAttribute("type1", "target")
	self:SetAttribute("ctrl-type1", "target")
	self:SetAttribute("toggleForVehicle", true) -- Vehicle Support
	header:CallMethod("InitializeUnitConfiguration", self:GetName())
    RegisterUnitWatch(self)
]]

-- Create a secure group header for each raid group
local groupFilters = { "1", "2", "3", "4", "5", "6", "7", "8" }
local groupHeaders = {}
local function CreateGroupHeader(pt, layout, index, isPlayer)
	local gname = (isPlayer and "Puffin_Group" or "Puffin_TargetGroup") .. index
	local r = CreateFrame("Frame", gname, UIParent, "SecureGroupHeaderTemplate")
	groupHeaders[gname] = r
	if r then
		r:Hide()
		local v = isPlayer and pt.player or pt.target
		local px, py = pt.offsetX, pt.offsetY
		if px ~= 0 then px = px - v.width end; if py ~= 0 then py = py - v.height end
		r:SetAttribute("template", "SecureUnitButtonTemplate"); r:SetAttribute("templateType", "Button")
		r:SetAttribute("groupFilter", groupFilters[index]); r:SetAttribute("groupBy", "GROUP"); r:SetAttribute("groupingOrder", groupFilters[index])
		r:SetAttribute("showRaid", true); r:SetAttribute("showParty", false); r:SetAttribute("showSolo", false); r:SetAttribute("showPlayer", false)
		r:SetAttribute("point", pt.rowPoint); r:SetAttribute("columnAnchorPoint", pt.columnPoint)
		r:SetAttribute("unitsPerColumn", 5); r:SetAttribute("maxColumns", 1); r:SetAttribute("xoffset", px); r:SetAttribute("yoffset", py)
		r:SetAttribute("initial-width", v.width); r:SetAttribute("initial-height", v.height); r:SetAttribute("initial-scale", pt.scale or 1)
		r:SetAttribute("initialConfigFunction", CreateSecureUnit); r:SetAttribute("sortMethod", "INDEX"); r:SetAttribute("startingIndex", 1)
		if not isPlayer then r:SetAttribute("SecureUnitSuffix", "target") end -- used when initializing units for raid targets
		r.UseClique = v.clique -- only set up Clique on selected frames
		r.InitializeUnitConfiguration = InitializeUnitConfiguration -- initialization function
		r:Show()
	end
	return r
end

-- Position a group header based on the layout's anchor settings
local function PositionGroupHeader(pt, layout, index, isPlayer)
	local gname = (isPlayer and "Puffin_Group" or "Puffin_TargetGroup") .. index
	local r = groupHeaders[gname]
	if r then
		local v = isPlayer and pt.player or pt.target
		local anchor, dx, dy = pt.anchor and layout.anchors[pt.anchor] or nil, 0, 0
		local anchorPoint = anchor and anchor.anchorPoint or "CENTER"
		local relativePoint = anchor and anchor.relativePoint or "CENTER"
		if (type(anchor) == "table") and (not anchor.x or not anchor.y) then
			dx = (anchor[index].x or 0) + v.offsetX -- use table of anchor points for each group
			dy = (anchor[index].y or 0) + v.offsetY
		else
			dx = (anchor and anchor.x or 0) + v.offsetX + ((index - 1) * pt.groupOffsetX) -- compute offsets from anchor to each group
			dy = (anchor and anchor.y or 0) + v.offsetY + ((index - 1) * pt.groupOffsetY)
		end
		r:SetPoint(anchorPoint, UIParent, relativePoint, dx, dy)
	end
end

-- On first update, configure the unit frames, afterwards just update each unit frame as needed
function Adapter_UnitFrames.OnUpdate()
	local inRaid, raidMembers, inParty = IsInRaid(), 0, false
	if inRaid then raidMembers = GetNumGroupMembers() else inParty = (GetNumGroupMembers() > 0) end
	
	local layout = DB.GetRecord("layouts", "standard")
	local index = GetSpecialization()
	if index then local _, spec = GetSpecializationInfo(index) if spec then clicks_record = DB.GetRecord("clicks", spec) end end
	
	if not Puffin.inLockdown and configuration_changed then -- can not configure while in combat lockdown
		if not units_configured then -- only configure primary units once, they will always be active
			units_configured = true
			for unit, v in pairs(layout.units) do
				local u = UF.Create() -- create the unit frame view
				u.unit = unit; u.defaults = layout.defaults; u.templates = v.templates or layout.templates; u.widgets = v.widgets
				u.height = v.height; u.width = v.width; u.depth = v.depth; u.clique = v.clique
				local anchor = v.anchor and layout.anchors[v.anchor] or nil
				u.offsetX = v.offsetX + (anchor and anchor.x or 0); u.offsetY = v.offsetY + (anchor and anchor.y or 0)
				u.anchorPoint = (anchor and anchor.anchorPoint or nil); u.relativePoint = (anchor and anchor.relativePoint or nil)
				UF.Configure(u)
				active_units[unit] = u
				unit_changed[unit] = true
				Puffin.RegisterCallback(UN.callbacks[unit], UnitChanged)
			end
			if layout.panels then
				pt = layout.panels
				for dname, v in pairs(pt) do ConfigurePanel(v, layout, v, dname) end
			end
		end
		
		local pt = layout.party
		if pt then
			if ((configuration_mode == "party") or (inParty and not inRaid)) then -- first check if in a party and need to configure the party units
				if not party_configured then -- only allocate party units if not in a raid too
					if pt.backdrop then ConfigurePanel(pt, layout, pt.backdrop, "party") end
					if pt.player then ConfigurePartyUnit(pt, layout, 1, true) end -- if showing player then always is first
					for i = 1, 4 do ConfigurePartyUnit(pt, layout, i, false) end
					party_configured = true
				end
			elseif party_configured then -- not in a party so must release any allocated party units
				ReleasePanel("party")
				if pt.player then ReleasePartyPlayer() end
				for i = 1, 4 do ReleasePartyUnit(i) end
				party_configured = false
			end
		end

		pt = layout.raid
		if pt then
			if not raid_configured then -- initialize raid group headers if there is a raid layout so available when needed
				if pt.player then for i = 1, raid_groups_max do CreateGroupHeader(pt, layout, i, true) end end
				if pt.target then for i = 1, raid_groups_max do CreateGroupHeader(pt, layout, i, false) end end
				if pt.backdrop then ConfigurePanel(pt, layout, pt.backdrop, "raid") end
				raid_configured = true
			end
			local raidGroups = 0 -- opportunity to move anchors based on changed number of raid groups...
			if inRaid then
				for i = 1, raidMembers do local _, _, sg = GetRaidRosterInfo(i); if sg > raidGroups then raidGroups = sg end end
			end
			if not raid_groups or (raidGroups ~= raid_groups) then
				raid_groups = raidGroups
				if pt.player then for i = 1, raid_groups_max do PositionGroupHeader(pt, layout, i, true) end end
				if pt.target then for i = 1, raid_groups_max do PositionGroupHeader(pt, layout, i, false) end end
			end
		end
	end

	if raid_configured and inRaid and configuration_changed and not configuration_updated then
		local pt = layout.raid
		if raidMembers > raid_members_configured then -- make sure have enough raid units configured
			for i = raid_members_configured + 1, raidMembers do ConfigureRaidUnit(pt, layout, i) end
			raid_members_configured = raidMembers
		end

		local raidGroups = 0; table.wipe(raid_members_group); table.wipe(raid_members_index)
		for i = 1, raidMembers do -- first figure out which group each raid member is in plus the total number of groups
			local _, _, sg = GetRaidRosterInfo(i)
			raid_members_group[i] = sg; if sg > raidGroups then raidGroups = sg end
		end
		if raidGroups > raid_groups_max then raidGroups = raid_groups_max end
		for i = 1, raidGroups do -- next figure out which index in each group is assigned to each raid member
			for u = 1, 5 do
				local b = _G["Puffin_Group" .. i .. "UnitButton" .. u]
				if b then
					local unitid, id = b:GetAttribute("unit"), nil
					if unitid then id = string.gsub(unitid, "raid", "") end
					if id and tonumber(id) then raid_members_index[tonumber(id)] = u end
				end
			end
		end
		
		for i = 1, raidMembers do -- position each raid member in the correct group
			local group, index = raid_members_group[i], raid_members_index[i] or 1
			if group and index then PositionRaidUnit(pt, layout, i, group, index - 1) end
		end
		
		local dp = active_panels.raid -- adjust panel size to fit around the current number of groups
		if dp then
			dp.height = pt.backdrop.height + math.abs(pt.groupOffsetY * raidGroups) - (pt.groupOffsetY - pt.player.height) - 2 -- adjust panel dimensions
			dp.width = pt.backdrop.width + math.abs(pt.groupOffsetX * raidGroups) -- allows it to grow either up/down or left/right
			DP.Configure(dp)
		end
	elseif raid_configured and not inRaid then -- not in a raid so release any allocated raid members
		ReleasePanel("raid")
		for i = 1, raid_members_configured do ReleaseRaidUnit(i) end
		raid_members_configured = 0
		raid_configured = false
	end
	
	if not Puffin.inLockdown then configuration_changed = false end -- only can reset this out of combat
	configuration_updated = true -- this can be reset at any time

	for dname, dp in pairs(active_panels) do -- update all display panels
		if dname == "party" then dp.hidden = not (configuration_mode or (inParty and not inRaid)) end
		if dname == "raid" then dp.hidden = not (configuration_mode or inRaid) end
		local record = nil
		if dp.unit then record = DB.GetRecord("units", dp.unit) end
		DP.Update(dp, record, dp.unit)
	end

	for unit, uf in pairs(active_units) do -- update all the unit frames that actually exist, hide the ones that don't
		local uid = uf.unit
		local t = DB.GetRecord("units", uid)
		if t then
			if t.inVehicle and t.vehicle then t = DB.GetRecord("units", t.vehicle)
			elseif unit == "pet" and t.isPossessed and t.possessor then t = DB.GetRecord("units", t.possessor) end
		end
		if configuration_mode then if not t then t = DB.GetRecord("units", "target"); unit = "target"; uid = "target" end unit_changed.target = true end
		uf.hidden = not UnitExists(uid) or not t or (uf.raidGroup and (uf.raidGroup > raid_groups_max)) -- hide raid members in groups not being displayed
		UF.Update(uf, unit_changed[uid], t, uid)
	end
	
	for unit, uf in pairs(active_units) do unit_changed[uf.unit] = false end
end

Puffin.RegisterComponent("Adapter_UnitFrames", Adapter_UnitFrames, "ADAPTER", 2)
