-- Puffin: A Tojaso FowlPlay Project
-- Model component for tracking unit-related information and triggering events for changes

-- Pending: threat, isBoss, alternate power info

local Model_Units = {}
local EF = Puffin.ErrorFormat
local DB, AU
local now = 0
local us = {}
local marked = {}
local changed = false
local libBossIDs = nil
local playerGUID, playerFaction

-- Tables of standard unit names
local standardUnits = { "player", "pet", "target", "focus", "targettarget", "focustarget", "pettarget", "mouseover", }
local partyUnits, partyTargets, raidUnits, raidTargets = {}, {}, {}, {}
for i = 1, 5 do partyUnits[i] = "party" .. i; partyTargets[i] = "party" .. i .. "target" end	
for i = 1, 40 do raidUnits[i] = "raid" .. i; raidTargets[i] = "raid" .. i .. "target" end

-- Table of callback names
Model_Units.callbacks = {}
local unitSets = { standardUnits, partyUnits, partyTargets, raidUnits, raidTargets }
for _, s in pairs(unitSets) do for _, unit in pairs(s) do Model_Units.callbacks[unit] = "units" .. "." .. unit end end

-- Check if there is a raid icon on a unit
local raidIcons = {} -- table of raid targets indexed by GUID
local function CheckRaidTarget(unit)
	if not unit then return end
	local id = UnitGUID(unit)
	if id then
		local index = GetRaidTargetIndex(unit)
		for k, v in pairs(raidIcons) do if (v == id) and (k ~= index) then raidIcons[k] = nil end end 
		if index then raidIcons[index] = id end
	end
end

-- Check raid targets on all addressable units
local function CheckRaidIcons()
	for _, unit in pairs(standardUnits) do CheckRaidTarget(unit) end -- first check primary units
	if not IsInRaid() then
		for i = 1, GetNumGroupMembers() do CheckRaidTarget(partyUnits[i]); CheckRaidTarget(partyTargets[i]) end
	else
		for i = 1, GetNumGroupMembers() do CheckRaidTarget(raidUnits[i]); CheckRaidTarget(raidTargets[i]) end
	end
	updateAll = true
end

-- Return the raid target index for a unit
local function GetRaidTarget(unit) local id = UnitGUID(unit) for k, v in pairs(raidIcons) do if v == id then return k end end return nil end

-- Range checking spell tables and functions
local range_spells = {}
local range_tables = {
	enemy = { DEATHKNIGHT = { 49576, 77606 }, DRUID = { 5176, 339, 6795, 770 }, HUNTER = { 75 }, MAGE = { 133 }, MONK = { 115546 }, PALADIN = { 62124 },
		PRIEST = { 589 }, ROGUE = { 1725 }, SHAMAN = { 403 }, WARLOCK = { 348, 686 }, WARRIOR = { 355 } },
	friend = { DEATHKNIGHT = { 49016 }, DRUID = { 50464, 467 }, MAGE = { 475 },  MONK = { 115921, 115450 }, PALADIN = { 635 }, HUNTER = { 13159 },
		PRIEST = { 2061 }, ROGUE = { 57934 }, SHAMAN = { 331 }, WARLOCK = { 5697 }, WARRIOR = { 3411 } },
	pet = { HUNTER = { 136 }, WARLOCK = { 755 } },
	dead = { DEATHKNIGHT = { 61999 }, DRUID = { 50769, 20484 }, MONK = { 115178 }, PALADIN = { 7328 }, PRIEST = { 2006 }, SHAMAN = { 2008 } },
}
-- Initialize the range spell tables with spell names
local function InitializeRangeSpells()
	local _, class = UnitClass("player")
	for k, v in pairs(range_tables) do
		if v[class] then
			local t = {}
			for _, sid in pairs(v[class]) do local name = GetSpellInfo(sid); if name then table.insert(t, name) end end
			range_spells[k] = t
		end
	end
end

-- Check a table to see if a unit is in range of any of the spells it contains
local function InRangeSpells(unit, t) if t then for _, s in pairs(t) do if IsSpellInRange(s, unit) == 1 then return true end end end return false end

-- Return true or false if the unit is in range based on whether it is enemy, friend or pet
local function UnitRangeCheck(unit, isPet, isDead)
	if not UnitExists(unit) or not UnitIsVisible(unit) then return false end -- check if unit is close enough for other tests to work properly
	if CheckInteractDistance(unit, 1) then return true end
	if UnitCanAttack("player", unit) then -- enemy unit
		if InRangeSpells(unit, range_spells.enemy) then return true end
	else -- friendly unit
		if isDead then return InRangeSpells(unit, range_spells.dead) end
		if InRangeSpells(unit, range_spells.friend) then return true end
		if isPet and InRangeSpells(unit, range_spells.pet) then return true end
	end
	return false
end

-- Get the direction to the unit under certain circumstances
local function GetDirection(t)
	if not t.inRange and t.isMouseover and t.isPlayer and t.isConnected and not WorldMapFrame:IsShown() then
		local x, y = GetPlayerMapPosition("player")
		if x == 0 and y == 0 then SetMapToCurrentZone(); x, y = GetPlayerMapPosition("player") end
		if x > 0 or y > 0 then
			local dx, dy = GetPlayerMapPosition("mouseover")
			if dx > 0 or dy > 0 then
				local facing = GetPlayerFacing()
				if facing < 0 then facing = facing + (2 * math.pi) end
				local theta = math.pi - math.atan2(x - dx, dy - y) - facing
				Puffin.StartTimer("Direction Update", 0.09) -- triggers update approximately ten times a second
				Puffin.mouseoverDirection = theta
				return theta
			end
		end
	end
	return nil
end

-- Track ready check status and responses
local readyChecking, readyCheck = false, {}
local function ReadyCheck() table.wipe(readyCheck); readyChecking = true end
local function ReadyCheckClear() table.wipe(readyCheck); readyChecking = false end
local function ReadyCheckConfirm(e, unit, response) local g = UnitGUID(unit); if g then readyCheck[g] = response and "ready" or "notready" end end
local function ReadyCheckFinished() Puffin.StartTimer("ReadyCheckClear", 1) end -- delay end by one second

-- Initialize by getting a handle on the database and registering any callbacks needed for unit-related game events
function Model_Units.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("Puffin.Model_Units", "data management component not found"))
	AU = Puffin.Component("Model_Auras") -- link to data management functions
	assert(AU, EF("Puffin.Model_Units", "aura component not found"))
	DB.AllocateDatabase("units", false)
	LoadAddOn("LibBossIDs-1.0")
	libBossIDs = LibStub("LibBossIDs-1.0", true)
	
	Puffin.RegisterCallback("RAID_TARGET_UPDATE", CheckRaidIcons)
	CheckRaidIcons() -- initialize raid target detection
	InitializeRangeSpells() -- initialize the range spell tables
	playerGUID = UnitGUID("player")
	playerFaction = UnitFactionGroup("player")
	
	local _
	_, Puffin.myClass = UnitClass("player") -- cache the player's class
	_, Puffin.myRace = UnitRace("player") -- cache the player's race

	Puffin.RegisterCallback("READY_CHECK", ReadyCheck)
	Puffin.RegisterCallback("READY_CHECK_CONFIRM", ReadyCheckConfirm)
	Puffin.RegisterCallback("READY_CHECK_FINISHED", ReadyCheckFinished)
	Puffin.RegisterCallback("ReadyCheckClear", ReadyCheckClear)
	
	Puffin.RegisterCallback("PLAYER_TARGET_CHANGED", Puffin.ForceUpdate)
	Puffin.RegisterCallback("PLAYER_FOCUS_CHANGED", Puffin.ForceUpdate)

	Puffin.RegisterCallback("UNIT_CHANGED", "null") -- register "null" callbacks for events that should cause updates
	Puffin.RegisterCallback("UNIT_HEALTH", "null")
	Puffin.RegisterCallback("UNIT_POWER", "null")
	Puffin.RegisterCallback("UNIT_COMBO_POINTS", "null")
	Puffin.RegisterCallback("UNIT_NAME_UPDATE", "null")
	Puffin.RegisterCallback("UNIT_CLASSIFICATION_CHANGED", "null")
	Puffin.RegisterCallback("UNIT_ENTERED_VEHICLE", "null")
	Puffin.RegisterCallback("UNIT_EXITED_VEHICLE", "null")
	Puffin.RegisterCallback("UNIT_FACTION", "null")
	Puffin.RegisterCallback("UNIT_FLAGS", "null")
	Puffin.RegisterCallback("UNIT_DYNAMIC_FLAGS", "null")	
	Puffin.RegisterCallback("UNIT_THREAT_SITUATION_UPDATE", "null")
	Puffin.RegisterCallback("COMBAT_LOG_EVENT_UNFILTERED", "null")
	Puffin.RegisterCallback("GROUP_ROSTER_UPDATE", "null")
	Puffin.RegisterCallback("PARTY_LEADER_CHANGED", "null")
	Puffin.RegisterCallback("PARTY_LOOT_METHOD_CHANGED", "null")
	Puffin.RegisterCallback("GUILD_ROSTER_UPDATE", "null")
	Puffin.RegisterCallback("UPDATE_MOUSEOVER_UNIT", "null")
	Puffin.RegisterCallback("UNIT_PET", "null")
end

-- Check a unit to see if it needs to be updated and if it has changed since previous update
-- Track time between unit updates and only do less volatile updates every 0.5 seconds
local function UpdateUnitRecord(unit, possessor)
	local u = nil -- this will be set to the unit record, if unit exists
	local callback = Model_Units.callbacks[unit] -- check if anyone has registered for info about this unit
	if not Puffin.CallbackRegistered(callback) then return end -- return if no one cares about this unit
	local g = UnitGUID(unit) -- multiple units can be the same actual unit, guid is unique per actual unit
	if g then -- check if the unit actually exists
		marked[g] = true -- mark so garbage collection recognizes the unit state table for this GUID is still active
		u = us[g]; if not u then u = Puffin.AllocateTable(); us[g] = u; u.guid = g; u.auras = AU.Allocate() end
		if u._updated ~= now then -- only need to update once per OnUpdate call
			local player, changed = (g == playerGUID), false
			local v = (g == UnitGUID("mouseover")); if u.isMouseover ~= v then u.isMouseover = v; changed = true end
			if not u._major or (now >= (u._major + 0.5)) then -- check if time for a major update
				local pet, inParty, inRaid = (g == UnitGUID("pet")), UnitInParty(unit), UnitInRaid(unit)
				local isMasterLooter, rank, raidGroup,raidZone, role, _
				v = UnitName(unit); if u.name ~= v then u.name = v; changed = true end
				if inRaid then
					_, rank, raidGroup, _, _, _, raidZone, _, _, role, isMasterLooter = GetRaidRosterInfo(inRaid)
					inRaid = (inRaid ~= nil); isMasterLooter = (isMasterLooter == 1)
					v = (rank == 2); if u.isRaidLeader ~= v then u.isRaidLeader = v; changed = true end
					v = (rank == 1); if u.isRaidAssistant ~= v then u.isRaidAssistant = v; changed = true end
					v = (role == "MAINASSIST"); if u.isMainAssist ~= v then u.isMainAssist = v; changed = true end
					v = (role == "MAINTANK"); if u.isMainTank ~= v then u.isMainTank = v; changed = true end
					if u.raidGroup ~= raidGroup then u.raidGroup = raidGroup; changed = true end
					if player and (u.raidZone ~= raidZone) then u.raidZone = raidZone; changed = true end
				else
					if inParty then
						_, v = GetLootMethod()
						isMasterLooter = (v ~= nil) and ((player and (v == 0)) or (not player and (g == UnitGUID("party" .. v))))
						v = UnitIsGroupLeader(unit); if u.isPartyLeader ~= v then u.isPartyLeader = v; changed = true end
					else
						isMasterLooter = false
					end
				end
				v = UnitLevel(unit); if u.level ~= v then u.level = v; changed = true end
				v = UnitIsConnected(unit); if u.isConnected ~= v then u.isConnected = v; changed = true end
				local dead = UnitIsDeadOrGhost(unit); if u.isDead ~= dead then u.isDead = dead; changed = true end
				v = UnitIsGhost(unit); if u.isGhost ~= v then u.isGhost = v; changed = true end
				if u.inParty ~= inParty then u.inParty = inParty; changed = true end
				if u.inRaid ~= inRaid then u.inRaid = inRaid; changed = true end
				if u.isMasterLooter ~= isMasterLooter then u.isMasterLooter = isMasterLooter; changed = true end
				local class, classToken = UnitClass(unit); if u.class ~= class then u.class = class; changed = true end
				if u.classToken ~= classToken then u.classToken = classToken; changed = true end
				v = UnitClassBase(unit); if u.classBase ~= v then u.classBase = v; changed = true end
				v = UnitRace(unit); if u.race ~= v then u.race = v; changed = true end
				v = UnitSex(unit); if u.sex ~= v then u.sex = v; changed = true end
				v = UnitIsEnemy("player", unit); if u.isEnemy ~= v then u.isEnemy = v; changed = true end	
				v = UnitIsFriend("player", unit); if u.isFriend ~= v then u.isFriend = v; changed = true end
				if libBossIDs and libBossIDs.BossIDs[tonumber(g:sub(-13, -9), 16)] then v = "worldboss" else v = UnitClassification(unit) end
				if u.classification ~= v then u.classification = v; changed = true end -- use LibBossIDs for better boss detection
				v = UnitIsPVP(unit); if u.pvp ~= v then u.pvp = v; changed = true end
				v = UnitFactionGroup(unit); if u.pvp == 1 and not v then v = playerFaction end if u.faction ~= v then u.faction = v; changed = true end
				v = UnitIsPlayer(unit); if u.isPlayer ~= v then u.isPlayer = v; changed = true end
				if v then -- these apply only to player units
					local guild, rank = GetGuildInfo(unit); if u.guild ~= guild then u.guild = guild; changed = true end
					if u.guildRank ~= rank then u.guildRank = rank; changed = true end
					v = UnitIsAFK(unit); if u.isAFK ~= v then u.isAFK = v; changed = true end
					v = UnitIsDND(unit); if u.isDND ~= v then u.isDND = v; changed = true end
					v = UnitIsFeignDeath(unit); if u.isFeign ~= v then u.isFeign = v; changed = true end
					v = UnitHasIncomingResurrection(unit) if u.resurrection ~= v then u.resurrection = v; changed = true end
					v = UnitIsPVPFreeForAll(unit); if u.pvpFFA ~= v then u.pvpFFA = v; changed = true end
					v = UnitGroupRolesAssigned(unit); if u.role ~= v then u.role = v; changed = true end
					v = UnitHasVehicleUI(unit); if u.inVehicle ~= v then u.inVehicle = v; changed = true end
					v = GetReadyCheckStatus(unit); if v then readyCheck[g] = v end
					if readyChecking then v = readyCheck[g] else v = nil end; if u.ready ~= v then u.ready = v; changed = true end
				else -- these apply to non-player units
					v = UnitCreatureFamily(unit); if u.creatureFamily ~= v then u.creatureFamily = v; changed = true end
					v = UnitCreatureType(unit); if u.creatureType ~= v then u.creatureType = v; changed = true end				
					v = UnitIsQuestBoss(unit); if u.isQuest ~= v then u.isQuest = v; changed = true end
					v = UnitIsTapped(unit); if u.isTapped ~= v then u.isTapped = v; changed = true end
					v = UnitIsTappedByPlayer(unit); if u.isTappedByPlayer ~= v then u.isTappedByPlayer = v; changed = true end
				end
				if player then -- this unit is the player
					u.inRange = true; u.direction = 0 -- player is always in range
					local inGroup = GetNumGroupMembers() > 0
					v = inGroup and IsInRaid(); if u.raid ~= v then u.raid = v; changed = true end
					v = inGroup and not IsInRaid(); if u.party ~= v then u.party = v; changed = true end
					v = IsResting(); if u.isResting ~= v then u.isResting = v; changed = true end
					local timer = GetPVPTimer(); v = (timer and timer > 0 and timer < 300000) and (timer / 1000) or nil
					if u.pvpTimer ~= v then u.pvpTimer = v; changed = true end
					if Puffin.myClass == "MONK" then v = UnitPower("player", SPELL_POWER_LIGHT_FORCE); if u.chiPoints ~= v then u.chiPoints = v; changed = true end end
				else -- this is not the player
					v = UnitRangeCheck(unit, pet, dead); if (u.inRange == true) ~= v then u.inRange = v; changed = true end
				end
				u._major = now
			end
			v = UnitHealth(unit); if u.health ~= v then u.health = v; changed = true end
			v = UnitHealthMax(unit); if u.healthMax ~= v then u.healthMax = v; changed = true end
			v = UnitGetIncomingHeals(unit); if u.healthIncoming ~= v then u.healthIncoming = v; changed = true end
			v = UnitPower(unit); if u.power ~= v then u.power = v; changed = true end
			v = UnitPowerMax(unit); if u.powerMax ~= v then u.powerMax = v; changed = true end
			v = UnitPowerType(unit); if u.powerType ~= v then u.powerType = v; changed = true end
			v = UnitAffectingCombat(unit); if u.inCombat ~= v then u.inCombat = v; changed = true end
			v = UnitIsCharmed(unit); if u.isCharmed ~= v then u.isCharmed = v; changed = true end
			v = UnitIsControlling(unit); if u.isControlling ~= v then u.isControlling = v; changed = true end
			v = UnitIsPossessed(unit); if u.isPossessed ~= v then u.isPossessed = v; changed = true end; u.possessor = nil
			v = GetRaidTarget(unit); if u.raidTarget ~= v then u.raidTarget = v; changed = true end
			v = UnitThreatSituation(unit); if u.threatSituation ~= v then u.threatSituation = v; changed = true end
			v = GetDirection(u); if u.direction ~= v then u.direction = v; changed = true end
			if not player then v = GetComboPoints("player", unit); if u.comboPoints ~= v then u.comboPoints = v; changed = true end end
			if AU.Auras(unit, u.auras) then u._auras_changed = true; changed = true end -- get current auras for the unit, check if changed
			u._updated = now; u._changed = changed -- keep track of whether any field actually changes during this update		
		end

		local newUnit = (g ~= DB.GetAttribute("units", unit, "guid")) -- compare previous GUID to see if it is for same unit
		if newUnit then
			DB.SetRedirect("units", unit, u) -- make sure record is set to current GUID
			Puffin.Callback(callback, unit, "new") -- callback with parameter indicating the unit has changed
		elseif u._changed then
			if u._auras_changed then Puffin.Callback(callback, unit, "auras") end -- callback when auras changed on the unit
			Puffin.Callback(callback, unit, "changed") -- callback indicating any info for current unit has changed
			if possessor then Puffin.Callback(callback, possessor, "possess") end -- also trigger callback for possessing unit
		end
	else
		if DB.ReleaseRecord("units", unit) then Puffin.Callback(callback, unit, "released") end -- callback indicating unit no longer exists
	end
	return u -- return the unit record, if the unit exists
end

-- Update a unit and, if it is a player unit and in a vehicle, also update the corresponding pet
local function UpdateUnit(unit)
	local u = UpdateUnitRecord(unit, nil)
	if u and u.inVehicle then
		local vehicle
		if unit == "player" then vehicle = "pet" elseif unit == "pet" then vehicle = "player" else vehicle = unit .. "pet" end
		local v = UpdateUnitRecord(vehicle, unit)
		if v then u.vehicle = vehicle; v.possessor = unit else u.vehicle = nil; u.inVehicle = nil end
	end
end

-- Update the database with current settings for all changed units
function Model_Units.OnUpdate()
	now = GetTime() -- track update times so that don't repeat updates when multiple units are actually the same GUID
	table.wipe(marked) -- initialize mark/sweep garbage collection of unit state tables
	
	if IsInRaid() then
		for i = 1, GetNumGroupMembers() do UpdateUnit(raidUnits[i]); UpdateUnit(raidTargets[i]) end -- update raid units first
	else
		for i = 1, GetNumGroupMembers() do UpdateUnit(partyUnits[i]); UpdateUnit(partyTargets[i]) end -- update party units second
	end
	for _, unit in pairs(standardUnits) do UpdateUnit(unit) end -- update primary units	last so raid and party info is already filled in if available
	
	if not Puffin.inLockdown then -- no garbage collection of unit state tables while in global lockdown
		for g, u in pairs(us) do if not marked[g] then u.auras = AU.Release(u.auras); us[g] = Puffin.ReleaseTable(u) end end -- release inactive
	end
end

-- Check a unit to see if it has changed since the last update
local function CheckUnit(unit)
	local callback = Model_Units.callbacks[unit] -- check if anyone has registered for info about this unit
	if not Puffin.CallbackRegistered(callback) then return end -- return if no one cares about this unit
	local g = UnitGUID(unit) -- get current guid for this unit and nil if doesn't exist
	if (g ~= DB.GetAttribute("units", unit, "guid")) then changed = true end -- compare to previous GUID to see if for same unit
end

-- Determine if any units that anyone cares about have changed and, if so, trigger an event
-- More responsive than waiting for UNIT_TARGET event which can be delayed by up to half a second
function Model_Units.OnRefresh()
	changed = false -- set to true as side effect when checking each unit
	for _, unit in pairs(standardUnits) do CheckUnit(unit) end -- check primary units
	if not IsInRaid() then
		for i = 1, GetNumGroupMembers() do CheckUnit(partyUnits[i]); CheckUnit(partyTargets[i]) end -- check party units, if any
	else
		for i = 1, GetNumGroupMembers() do CheckUnit(raidUnits[i]); CheckUnit(raidTargets[i]) end -- check raid units, if any
	end
	if changed then Puffin.Callback("UNIT_CHANGED") end -- trigger special event

	local g = UnitGUID("mouseover"); if g then -- handle mouseover direction as special case for animations, set global with current value
		local u = us[g]; if u then GetDirection(u) end -- has side effect of setting Puffin.mouseoverDirection, used for icon animations
	end
end

Puffin.RegisterComponent("Model_Units", Model_Units, "MODEL", 0)
