-- Puffin: A Tojaso FowlPlay Project
-- Action button widget for interactive icons (e.g., data brokers, icons for timer bars)
-- Supports icon texture, cooldown overlay, border color, start/end animations
-- Includes default border texture but supports optional use of Masque to fully skin button
-- Icon are assumed square , interactive container frame width and height set separately

-- Widget attributes:
-- texture - path to button's icon's texture
-- texcoord - table containing icon's texcoord settings (default: { 0, 1, 0, 1 })
-- iconSize - width and height of the button's icon (assumed to be square)
-- width - width of the button's interactive container frame
-- height - height of the button's interactive container frame
-- color - color to apply to button's icon (default: none)
-- clockDuration - clock overlay duration, valid if > 0 (default: no overlay)
-- clockExpire - clock overlay expire time (only if clockDuration > 0)
-- clockDirection - if true show clock overlay in reverse direction (default: normal direction)
-- masque - enable use of Masque addon if it is installed (default: false)
-- masqueColor - color to apply to Masque border (default: none)
-- onClick - function to call when button click is detected
-- onEnter - function to call when mouse enters button
-- onLeave - function to call when mouse leaves button
-- eventHandler - table that is passed to mouse interaction functions (created by the widget)
-- scale - factor to scale widget's display size
-- depth - relative to view's depth (range: 0 .. 9)
-- alpha - opacity for the widget
-- hidden - hide the widget

local Widget_Button = {}
local EF = Puffin.ErrorFormat
local WA, DB
local recycledButtons = {}
local buttonCount = 0
local MSQ = nil
local MSQ_ButtonData = nil

-- Create a new button widget object
function Widget_Button.Create()
	local n, b = next(recycledButtons) -- get any available recycled bar
	if n then
		recycledButtons[n] = nil
	else
		b = {}
		buttonCount = buttonCount + 1
		b._display = {} -- table of display objects preserved when widget is recycled
		local t = b._display
		local frameName = "Puffin_Button" .. tostring(buttonCount) -- generate unique name for container frame
		t.container = CreateFrame("Button", frameName .. "Container")
		t.container:SetFrameStrata("MEDIUM")
		b.eventHandler = {}; t.container._eventHandler = b.eventHandler -- table to pass info to event handlers
		t.texture = t.container:CreateTexture(nil, "ARTWORK")
		t.cooldown = CreateFrame("Cooldown", frameName .. "Cooldown", t.container) -- cooldown overlay to animate timer
		t.cooldown.noCooldownCount = true -- disable OmniCC on button icons
		t.border = t.container:CreateTexture(nil, "BACKGROUND")
		if MSQ then
			t.buttonData = {} -- only initialize once so no garbage collection issues
			for k, v in pairs(MSQ_ButtonData) do t.buttonData[k] = v end
		end
	end
	return b
end

-- Release when no longer needed, recycle object after clearing attributes and hiding display objects associated with it
function Widget_Button:Release()
	if MSQ and self.masque then local mg = MSQ:Group("Puffin", self.masque); if mg then mg:RemoveButton(t.container) end end
	for k in pairs(self) do if k ~= "_display" and k ~= "eventHandler" then self[k] = nil end end
	local t = self._display; table.wipe(self.eventHandler)
	t.container:Hide(); t.texture:Hide(); t.cooldown:Hide(); t.textFrame:Hide(); t.text:Hide(); t.border:Hide()
	table.insert(recycledButtons, self)
end

-- Set display position for the widget relative to another widget or the overall screen
function Widget_Button:SetPosition(anchor, anchorWidget, relative, offsetX, offsetY, region)
	local t, w = self._display, UIParent
	if anchorWidget then
		if anchorWidget._display then w = region and anchorWidget._display[region] or anchorWidget._display.container else w = anchorWidget end
	end
	t.container:ClearAllPoints(); t.container:SetPoint(anchor, w, relative, offsetX, offsetY)
end

-- Set the parent widget or frame (used for managing display heirarchy)
function Widget_Button:SetParent(widget)
	local t, w = self._display, widget._display
	if w and w.container then w = w.container else w = widget end
	if not w then w = UIParent end
	t.container:SetParent(w)
	if not self.depth then t.container:SetFrameLevel(w:GetFrameLevel() + 1) end
end

-- Update depth for the widget
function Widget_Button:SetDepth(depth)
	local t = self._display; if not t then return end
	local b = t.container; if not b then return end
	self.depth = depth
	if self.depth then
		b:SetFrameLevel(self.depth)
		t.cooldown:SetFrameLevel(self.depth + 1)
	end
end

-- Functions to call when mouse interactions are detected
local function Button_OnEnter(b, ...) local f = b._onEnter if f then f(b, b._eventHandler, ...) end end
local function Button_OnLeave(b, ...) local f = b._onLeave if f then f(b, b._eventHandler, ...) end end
local function Button_OnClick(b, ...) local f = b._onClick if f then f(b, b._eventHandler, ...) end end

-- Configure attributes associated with the action button
function Widget_Button:Configure()
	assert(self.iconSize and self.iconSize > 0, EF("Widget_Button", "undefined iconSize"))
	assert(self.width and self.width >= self.iconSize, EF("Widget_Button", "undefined width"))
	assert(self.height and self.height >= self.iconSize, EF("Widget_Button", "undefined height"))
	local t = self._display; if not t then return end
	local b = t.container; if not b then return end
	t.texture:SetSize(self.iconSize, self.iconSize); t.texture:Hide()
	b:SetSize(self.iconSize, self.iconSize); b:SetScale(self.scale or 1); b:Hide()
	b._onEnter = self.onEnter; b._onLeave = self.onLeave; b._onClick = self.onClick; table.wipe(self.eventHandler)
	b:SetScript("OnEnter", Button_OnEnter); b:SetScript("OnLeave", Button_OnLeave); b:SetScript("OnClick", Button_OnClick)
	b:SetHitRectInsets(0, self.iconSize - self.width, self.iconSize - self.height, 0); b:RegisterForClicks("AnyUp")
	-- b.tex = b:CreateTexture(nil, "ARTWORK"); b.tex:SetAllPoints(); b.tex:SetTexture(0,1,0,0.5); b.tex:Show() -- debug
end

-- Get a handle for the optional animation widget
function Widget_Button.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("Widget_Button", "Data component not found"))
	WA = Puffin.Component("Widget_Animate") -- handle for optional animation widget class
	local masque = DB.GetAttribute("defaults", "Masque", "enable")
	if masque then
		MSQ = LibStub("Masque", true) -- MSQ is only non-nil when enabled and library loaded
		if MSQ then
			MSQ_ButtonData = {
				AutoCast = false, AutoCastable = false, Border = false, Checked = false, Cooldown = false, Count = false, Duration = false,
				Disabled = false, Flash = false, Highlight = false, HotKey = false, Icon = false, Name = false, Normal = false, Pushed = false
			}
		end
	end
end

-- Update function for button widget
function Widget_Button:Update()
	local t = self._display; if not t then return end
	local b = t.container; if not b then return end
	if self.hidden then
		t.texture:ClearAllPoints(); t.texture:Hide(); t.cooldown:Hide()
		b:Hide(); b:EnableMouse(false)
	else
		if self.texture then t.texture:SetTexture(self.texture) else t.texture:SetTexture(1, 1, 1, 0) end
		if self.texcoord then t.texture:SetTexCoord(unpack(self.texcoord)) else t.texture:SetTexCoord(0, 1, 0, 1) end
		local c = self.color; if c then t.texture:SetVertexColor(c.r, c.g, c.b) else t.texture:SetVertexColor(1, 1, 1) end
		t.texture:ClearAllPoints(); t.texture:SetPoint("BOTTOMLEFT", b, "BOTTOMLEFT"); t.texture:Show()
		
		if self.clockDuration and self.clockExpire and self.clockDuration > 0 then
			t.cooldown:SetCooldown(self.clockExpire - self.clockDuration, self.clockDuration)
			t.cooldown:SetReverse(self.clockDirection == true)
			t.cooldown:ClearAllPoints(); t.cooldown:SetPoint("CENTER", b, "CENTER"); t.cooldown:Show()
		else
			t.cooldown:SetCooldown(0, 0); t.cooldown:Hide()
		end

		if self.depth then
			b:SetFrameLevel(self.depth)
			t.cooldown:SetFrameLevel(self.depth + 1)
		end
		b:SetAlpha(self.alpha or 1); b:EnableMouse(true); b:Show()
		
		if self.texture and MSQ and self.masque then
			local mg = MSQ:Group("Puffin", self.masque)
			if mg then
				mg:RemoveButton(b, true) -- needed so size changes work when button is reused
				local bdata = t.buttonData
				bdata.Icon = t.texture
				bdata.Normal = b:GetNormalTexture()
				bdata.Cooldown = t.cooldown
				bdata.Border = t.border
				mg:AddButton(b, bdata)
				local cc = self.masqueColor
				if cc then t.border:SetVertexColor(cc.r, cc.g, cc.b, cc.a) end
			end
		end
	end
end

-- Register this widget as a component so it can be found by views
Puffin.RegisterComponent("Widget_Button", Widget_Button, "WIDGET", 0)
