-- Puffin: A Tojaso FowlPlay Project
-- Widget to conditionally show an icon

-- Custom attributes:
-- texture - path to icons's texture (default: color swatch) [d]
-- texcoord - table containing icon's texcoord settings (default: { 0, 1, 0, 1 }) [d]
-- color - color to apply to icon (default: none) [d]
-- condition - only show the icon when true (default: false) [d]
-- rotation - amount of rotation to apply to the texture in radians (default: none) [d]
-- startAnimation - animation to play when icon is shown: fade, pulse, delay, splash (default: none) [d]
-- endAnimation - animation to play when icon is hidden: fade, pulse, delay, splash (default: none) [d]

local Widget_Icon = {}
local EF = Puffin.ErrorFormat
local WA = nil -- optional animation widget
local recycledIcons = {}
local activeIcons = {}
local iconNumber = 0 -- used to create unique frame names

-- Create a new icon widget object
function Widget_Icon.Create()
	local n, icon = next(recycledIcons) -- get any available recycled widget
	if n then
		recycledIcons[n] = nil
	else
		icon = {}
		iconNumber = iconNumber + 1
		local t = {}; icon._display = t -- table of display objects preserved when widget is recycled
		local frameName = "Puffin_Icon" .. tostring(iconNumber) -- generate unique name for widget's frame
		t.container = CreateFrame("Frame", frameName .. "Container", UIParent)
		t.container:SetFrameStrata("MEDIUM")
		t.texture = t.container:CreateTexture(nil, "ARTWORK")
	end
	table.insert(activeIcons, icon) -- keep track of widgets so can animate if necessary
	return icon
end

-- Release widget when no longer needed, recycle object after clearing attributes and hiding display objects associated with it
function Widget_Icon:Release()
	for k in pairs(self) do if k ~= "_display" then self[k] = nil end end
	self._display.container:Hide()
	for k, v in pairs(activeIcons) do if v == self then activeIcons[k] = nil end end
	table.insert(recycledIcons, self)
end

-- Set display position for the widget relative to another widget or the overall screen
function Widget_Icon:SetPosition(anchor, anchorWidget, relative, offsetX, offsetY, region)
	local t, w = self._display, UIParent
	if anchorWidget then
		if anchorWidget._display then w = region and anchorWidget._display[region] or anchorWidget._display.container else w = anchorWidget end
	end
	t.container:ClearAllPoints(); t.container:SetPoint(anchor, w, relative, offsetX, offsetY)
end

-- Set the parent widget or frame (used for managing display heirarchy)
function Widget_Icon:SetParent(widget)
	local t, w = self._display, widget._display
	if w and w.container then w = w.container else w = widget end
	if not w then w = UIParent end
	t.container:SetParent(w)
	if not self.depth then t.container:SetFrameLevel(w:GetFrameLevel() + 1) end
end

-- Update depth for the widget
function Widget_Icon:SetDepth(depth)
	local t = self._display; if not t then return end
	local b = t.container; if not b then return end
	self.depth = depth
	if self.depth then b:SetFrameLevel(self.depth) end
end

-- Configure all the widget's display objects based on current configuration attribute values (or defaults)
function Widget_Icon:Configure()
	assert(self.width and self.width > 0, EF("Widget_Icon", "undefined width"))
	assert(self.height and self.height > 0, EF("Widget_Icon", "undefined height"))
	local w, h = self.width, self.height
	if w > h then w = h elseif h > w then h = w end -- make sure texture is square
	local t = self._display
	local tc = t.container
	t.texture:SetAlpha(1); t.texture:SetSize(w, h)
	tc:SetScale(self.scale or 1); tc:SetSize(self.width, self.height); 
	-- tc.tex = tc:CreateTexture(nil, "ARTWORK"); tc.tex:SetAllPoints(); tc.tex:SetTexture(0,0,1,0.5); tc.tex:Show() -- debug
	t.texture:Hide(); tc:Hide()
end

-- Get a handle for the optional animation widget
function Widget_Icon.OnEnable()
	WA = Puffin.Component("Widget_Animate") -- handle for optional animation widget class
end

-- Update the widget's display objects based on current data attribute values (or defaults)
function Widget_Icon:Update()
	local t = self._display
	if self.hidden or (self.condition == false) then
		if t.container:IsShown() and self.endAnimation and WA then WA.IconAnimation(self, self.endAnimation) end
		t.texture:ClearAllPoints(); t.texture:SetPoint("BOTTOMLEFT", t.container, "BOTTOMRIGHT"); t.texture:SetPoint("TOPRIGHT", t.container, "TOPLEFT")
		t.texture:Hide(); t.container:Hide() -- setting size to 0 supports automatically positioning rows of icons
	else
		if not t.container:IsShown() and self.startAnimation and WA then WA.IconAnimation(self, self.startAnimation) end
		if self.texture then t.texture:SetTexture(self.texture) else t.texture:SetTexture(1, 1, 1) end
		if self.texcoord then t.texture:SetTexCoord(unpack(self.texcoord)) else t.texture:SetTexCoord(0, 1, 0, 1) end
		t.texture:ClearAllPoints(); t.texture:SetPoint("BOTTOMLEFT", t.container, "BOTTOMLEFT")
		if self.rotation then t.texture:SetRotation(self.rotation); self._rotation = self.rotation elseif self._rotation then self._rotation = nil end
		
		local d = self.depth; if d then t.container:SetFrameLevel(d) end
		local c = self.color; if c then t.texture:SetVertexColor(c.r, c.g, c.b) else t.texture:SetVertexColor(1, 1, 1) end
		t.container:SetAlpha(self.alpha or 1); t.texture:Show(); t.container:Show()
	end
end

-- Special-case animation of rotations of the texture
function Widget_Icon:AnimateRotation()
	if not (self.hidden or (self.condition == false)) then
		local dir = Puffin.mouseoverDirection
		if dir and self._rotation and dir ~= self._rotation then
			local t = self._display
			if t.texture then t.texture:SetRotation(dir); self._rotation = dir end
		end
	end
end

function Widget_Icon:OnRefresh() for _, w in pairs(activeIcons) do Widget_Icon.AnimateRotation(w) end end

-- Register this widget as a component so it can be found by views
Puffin.RegisterComponent("Widget_Icon", Widget_Icon, "WIDGET", 1)
