-- Puffin: A Tojaso FowlPlay Project
-- Widget to display a panel

-- Custom attributes:
-- borderTexture - if defined then show a border around the panel [c]
-- borderColor - color to apply to border (default: black) [d]
-- borderWidth - width of border (default: 8) [c]
-- borderOffset - offset of border from texture (default: 2) [c]
-- borderOffsetX, borderOffsetY - position offset for border and panel (default: none) [c]
-- panelTexture - if defined then fill the panel with colored texture [c]
-- panelColor - color to apply to panel's fill area (default: white) [d]
-- panelInset - inset of panel texture from border (default: 0) [c]
-- panelAlpha - alpha adjustment for panel's opacity (default: 1) [d]

local Widget_Panel = {}
local EF = Puffin.ErrorFormat
local recycledPanels = {}
local panelNumber = 0
local defaultWhite = { r = 1, g = 1, b = 1, a = 1 }
local defaultBlack = { r = 0, g = 0, b = 0, a = 1 }
local borderTable = { tile = false, tileSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0 }}

-- Create a new panel widget object
function Widget_Panel.Create()
	local n, panel = next(recycledPanels) -- get any available recycled widget
	if n then
		recycledPanels[n] = nil
	else
		panel = {}
		panelNumber = panelNumber + 1
		local t = {}; panel._display = t -- table of display objects preserved when widget is recycled
		local frameName = "Puffin_Panel" .. tostring(panelNumber) -- generate unique name for frames
		t.container = CreateFrame("Frame", frameName .. "Panel", UIParent)
		t.container:SetFrameStrata("MEDIUM")
		t.border = CreateFrame("Frame", frameName .. "Border", t.container)
	end
	return panel
end

-- Release text when no longer needed, recycle object after clearing attributes and hiding display objects associated with it
function Widget_Panel:Release()
	for k in pairs(self) do if k ~= "_display" then self[k] = nil end end
	self._display.border:SetBackdrop(nil); self._display.border:Hide(); self._display.container:Hide()
	table.insert(recycledPanels, self)
end

-- Set display position for the widget relative to another widget or the overall screen
function Widget_Panel:SetPosition(anchor, anchorWidget, relative, offsetX, offsetY, region)
	local t, w = self._display, UIParent
	if anchorWidget then
		if anchorWidget._display then w = region and anchorWidget._display[region] or anchorWidget._display.container else w = anchorWidget end
	end
	t.container:ClearAllPoints(); t.container:SetPoint(anchor, w, relative, offsetX, offsetY)
end

-- Set the parent widget or frame (used for managing display heirarchy)
function Widget_Panel:SetParent(widget)
	local t, w = self._display, widget._display
	if w and w.container then w = w.container else w = widget end
	if not w then w = UIParent end
	t.container:SetParent(w)
	if not self.depth then local d = w:GetFrameLevel() + 1; t.container:SetFrameLevel(d); t.border:SetFrameLevel(d + 1) end
end

-- Configure all the widget's display objects based on current configuration attribute values (or defaults)
function Widget_Panel:Configure()
	local t = self._display
	assert(self.width and self.width > 0, EF("Widget_Panel", "undefined width"))
	assert(self.height and self.height > 0, EF("Widget_Panel", "undefined height"))
	assert(self.borderTexture or self.panelTexture, EF("Widget_Panel", "undefined texture"))
		
	t.container:SetSize(self.width, self.height); t.container:SetScale(self.scale or 1)	
	local offset, edgeSize = self.borderOffset or 2, self.borderWidth or 8; if (edgeSize < 0.1) then edgeSize = 0.1 end
	borderTable.bgFile = self.panelTexture
	local x, c = self.panelInset or 0, borderTable.insets; c.left = x; c.right = x; c.top = x; c.bottom = x
	borderTable.edgeFile = self.borderTexture; borderTable.edgeSize = edgeSize; t.border:SetBackdrop(borderTable)
	t.border:SetSize(self.width + offset, self.height + offset)
	t.border:ClearAllPoints(); t.border:SetPoint("CENTER", t.container, "CENTER", self.borderOffsetX or 0, self.borderOffsetY or 0)
	t.border:Hide(); t.container:Hide()
end

-- Update display objects based on current attribute values (or defaults)
function Widget_Panel:Update()
	local t = self._display
	if self.hidden or (self.condition == false) or not (self.borderTexture or self.panelTexture) then
		t.border:Hide(); t.container:Hide()
	else
		local d = self.depth; if d then t.container:SetFrameLevel(d); t.border:SetFrameLevel(d + 1) end
		local c = self.borderColor or defaultBlack; t.border:SetBackdropBorderColor(c.r, c.g, c.b, c.a)
		c = self.panelColor or defaultWhite; t.border:SetBackdropColor(c.r, c.g, c.b, self.panelAlpha or 1)
		t.container:SetAlpha(self.alpha or 1); t.border:Show(); t.container:Show()
	end
end

-- Register this widget as a component so it can be found by views
Puffin.RegisterComponent("Widget_Panel", Widget_Panel, "WIDGET", 0)
